/*
 * Copyright (c) 2014 Cloudera, Inc. All rights reserved.
 */

package com.cloudera.keytrustee;

import com.cloudera.keytrustee.util.ConfigUpdate;
import com.cloudera.keytrustee.util.Environment;
import com.google.common.base.Preconditions;
import org.apache.commons.io.FileUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.crypto.key.kms.server.KMSConfiguration;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.security.ProviderUtils;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.HashMap;
import java.util.List;
import java.util.Arrays;
import java.util.Map;

/**
 * Utility class to load KeyTrustee key provider configuration.
 * Based on org.apache.hadoop.crypto.key.kms.server.KMSConfiguration
 */
public class TrusteeKeyProviderConfiguration {
  public static final String KTS_CONFIG_DIR = "kts.config.dir";
  public static final String KTS_SITE_XML = "kts-site.xml";
  public static final String KTS_ACLS_XML = "kms-acls.xml";

  public static final String CONFIG_PREFIX = "cloudera.trustee.keyprovider.";
  public static final String ACL_CONFIG_PREFIX = "keytrustee.kms.";

  public static final String KEYTRUSTEE_CONF_DIR_ENV_VAR = "KEYTRUSTEE_CONF_DIR";
  public static final String KEYTRUSTEE_CONF_DIR_KEY = "hadoop.security.keystore.conf-dir";

  public static final String DEFAULT_KEYTRUSTEE_CONF_DIR_VALUE =
          "/var/lib/kms-keytrustee/keytrustee/.keytrustee";
  public static final String DEFAULT_PARCEL_LOCATION = "/opt/cloudera/parcels";
  public static final String DEFAULT_KEYTRUSTEE_55_PARCEL = "KEYTRUSTEE-5.5.0-1.cdh5.5.0.p0.215";
  
  public static final String KEYTRUSTEE_SERVER_ACTIVE_HOST_ENV_VAR = "KEYTRUSTEE_HOSTNAME_ACTIVE";
  public static final String KEYTRUSTEE_SERVER_PASSIVE_HOST_ENV_VAR = "KEYTRUSTEE_HOSTNAME_PASSIVE";
  public static final String KEYTRUSTEE_SERVER_ORG_ENV_VAR = "KEYTRUSTEE_ORG";
  public static final String KEYTRUSTEE_SERVER_AUTH_ENV_VAR = "KEYTRUSTEE_AUTH";
  public static final String KEYTRUSTEE_SERVER_CERT_FILE_ENV_VAR = "KEYTRUSTEE_CERT";
  public static final String KEYTRUSTEE_SERVER_KEY_LENGTH_ENV_VAR = "KEYTRUSTEE_KEYLENGTH";
  public static final String KEYTRUSTEE_SERVER_INSECURE_ENV_VAR = "KEYTRUSTEE_INSECURE";
  public static final String KEYTRUSTEE_SERVER_PROTOCOL_ENV_VAR = "KEYTRUSTEE_PROTOCOL";
  public static final String KEYTRUSTEE_SERVER_TOKEN_SYNC_ENV_VAR = "KEYTRUSTEE_TOKENSYNC";
  public static final String KEYTRUSTEE_SERVER_ROUND_ROBIN_ENV_VAR = "KEYTRUSTEE_ROUND_ROBIN";
  public static final String KEYTRUSTEE_SERVER_HKP_PORT_ENV_VAR = "KEYTRUSTEE_HKP_PORT";
  public static final String KEYTRUSTEE_SERVER_KTS_PORT_ENV_VAR = "KEYTRUSTEE_KTS_PORT";
  public static final String KEYTRUSTEE_SERVER_HKP_SSL_ENV_VAR = "KEYTRUSTEE_HKP_SSL";
  public static final String KEYTRUSTEE_CLIENT_USE_POOL_ENV_VAR = "KEYTRUSTEE_USE_POOL";
  public static final String KEYTRUSTEE_CLIENT_STRONG_LOCKING_ENV_VAR = "KEYTRUSTEE_STRONG_LOCKING";

  
  public static final String KEYTRUSTEE_SERVER_ACTIVE_HOST_KEY = "cloudera.trustee.keyprovider.hostname-ACTIVE";
  public static final String KEYTRUSTEE_SERVER_PASSIVE_HOST_KEY = "cloudera.trustee.keyprovider.hostname-PASSIVE";
  public static final String KEYTRUSTEE_SERVER_ORG_KEY = "cloudera.trustee.keyprovider.org";
  public static final String KEYTRUSTEE_SERVER_AUTH_KEY = "cloudera.trustee.keyprovider.auth";
  public static final String KEYTRUSTEE_SERVER_CERT_FILE_KEY = "cloudera.trustee.keyprovider.cert";
  public static final String KEYTRUSTEE_SERVER_KEY_LENGTH_KEY = "cloudera.trustee.keyprovider.keylength";
  public static final String KEYTRUSTEE_SERVER_INSECURE_KEY = "cloudera.trustee.keyprovider.insecure";
  public static final String KEYTRUSTEE_SERVER_PROTOCOL_KEY = "cloudera.trustee.keyprovider.protocol";
  public static final String KEYTRUSTEE_SERVER_TOKEN_SYNC_KEY = "cloudera.trustee.keyprovider.tokensync";
  public static final String KEYTRUSTEE_SERVER_ROUND_ROBIN_KEY = "cloudera.trustee.keyprovider.roundrobin";
  public static final String KEYTRUSTEE_SERVER_HKP_PORT_KEY = "cloudera.trustee.keyprovider.hkpport";
  public static final String KEYTRUSTEE_SERVER_KTS_PORT_KEY = "cloudera.trustee.keyprovider.ktsport";
  public static final String KEYTRUSTEE_SERVER_HKP_SSL_KEY = "cloudera.trustee.keyprovider.hkpssl";
  public static final String KEYTRUSTEE_CLIENT_USE_POOL_KEY = "cloudera.trustee.keyprovider.usepool";
  public static final String KEYTRUSTEE_CLIENT_STRONG_LOCKING_KEY = "cloudera.trustee.keyprovider.stronglocking";

  public static final String KEYTRUSTEE_CLIENT_POOL_MAX_SIZE_ENV_VAR = "KEYTRUSTEE_POOL_MAX";
  public static final String KEYTRUSTEE_CLIENT_POOL_MIN_IDLE_ENV_VAR = "KEYTRUSTEE_POOL_MIN_IDLE";
  public static final String KEYTRUSTEE_CLIENT_POOL_MAX_IDLE_ENV_VAR = "KEYTRUSTEE_POOL_MAX_IDLE";
  public static final String KEYTRUSTEE_CLIENT_POOL_ABANDONED_TIMEOUT_ENV_VAR = "KEYTRUSTEE_POOL_ABANDONED_TIMEOUT";

  public static final String KEYTRUSTEE_CLIENT_POOL_MAX_SIZE_KEY = "cloudera.trustee.keyprovider.pool.max";
  public static final String KEYTRUSTEE_CLIENT_POOL_MIN_IDLE_KEY = "cloudera.trustee.keyprovider.pool.min.idle";
  public static final String KEYTRUSTEE_CLIENT_POOL_MAX_IDLE_KEY = "cloudera.trustee.keyprovider.pool.max.idle";
  public static final String KEYTRUSTEE_CLIENT_POOL_ABANDONED_TIMEOUT_KEY = "cloudera.trustee.keyprovider.pool.abandoned.timeout";

  public static final String KEYTRUSTEE_CLIENT_READ_CONNECTION_TIMEOUT_ENV_VAR = "KEYTRUSTEE_HTTP_READ_TIMEOUT";
  public static final String KEYTRUSTEE_CLIENT_READ_CONNECTION_TIMEOUT_KEY = "cloudera.keytrustee.http.read.connectiontimeout";


  private static final Log LOG = LogFactory.getLog(TrusteeKeyProviderConfiguration.class);
  
  private Environment env;
  private Configuration conf;
  private URI[] uris;
  private File keyTrusteeConfDir;
  private URL keyTrusteeURL;

  static Configuration getConfiguration(boolean loadHadoopDefaults,
                                        String ... resources) {
    Configuration conf = new Configuration(loadHadoopDefaults);
    String confDir = System.getProperty(KTS_CONFIG_DIR);
    if (null == confDir) confDir = System.getProperty(KMSConfiguration.KMS_CONFIG_DIR);
    if (confDir != null) {
      try {
        if (!confDir.startsWith("/")) {
          throw new RuntimeException("System property '" + KTS_CONFIG_DIR +
                  "' must be an absolute path: " + confDir);
        }
        if (!confDir.endsWith("/")) {
          confDir += "/";
        }
        for (String resource : resources) {
          conf.addResource(new URL("file://" + confDir + resource));
          System.err.println(confDir);
        }
      } catch (MalformedURLException ex) {
        throw new RuntimeException(ex);
      }
    } else {
      for (String resource : resources) {
        conf.addResource(resource);
      }
    }
    return conf;
  }

  public static Configuration getKTSConf() {
    return getConfiguration(true, KTS_SITE_XML);
  }

  public static Configuration getACLsConf() { return getConfiguration(false, KTS_ACLS_XML); }
  
  public TrusteeKeyProviderConfiguration(Environment env, URI uri, Configuration conf) throws IOException {
    if (null == env) {
      this.env = new Environment();
    } else {
      this.env = env;
    }

    Preconditions.checkNotNull(conf, "Null configuration passed to constructor.");
    this.conf = new Configuration(conf);
    this.conf.addResource(getKTSConf());
    
    this.setKeyTrusteeConfDirAndURL(uri);
    this.moveDefaultConfigurationDir();
    
  }

  public File getKeyTrusteeConfDir() { return this.keyTrusteeConfDir; }
  public URL getKeyTrusteeURL() { return this.keyTrusteeURL; }
  public URI[] getSpecifiedProviderURIs() { return this.uris; }
  
  public int getKeyLength() {
    return getKeyLength(this.conf, this.env);
  }
  public boolean isUseKeyTrusteeConnectionPool() {
    return isUseKeyTrusteeConnectionPool(this.conf, this.env);
  }
  public boolean isStrongLocking() {
    return isStrongLocking(this.conf, this.env);
  }
  
  public Register getRegisterCommand() {
    return new Register(this.conf, this.env);
  }
  
  public int getMaxIdle() { return getMaxIdle(this.conf, this.env); }
  public int getMinIdle() { return getMinIdle(this.conf, this.env); }
  public int getMaxTotal() { return getMaxTotal(this.conf, this.env); }
  public int getAbandonedTimeout() { return getAbandonedTimeout(this.conf, this.env); }

  public int getHttpReadConnectionTimeout() { return getHttpReadConnectionTimeout(this.conf, this.env); }
  
  public boolean updateClientConfiguration(ClientInfo keyTrusteeClient) throws KeyTrusteeException, IOException {
    return updateClientConfiguration(keyTrusteeClient, false);
  }
  
  public boolean updateClientConfiguration(ClientInfo keyTrusteeClient, boolean forceUpdate) 
          throws KeyTrusteeException, IOException { 
    return updateClientConfiguration(this.conf, this.env, keyTrusteeClient, forceUpdate);
  }
  
/*
 *
 */
  public static boolean updateClientConfiguration(Configuration conf, Environment env, ClientInfo keyTrusteeClient, boolean forceUpdate)
          throws KeyTrusteeException, IOException {
    
    String targetMasterHostname = getMasterHostname(conf, env);

    FailoverServerInfo keyTrusteeServer = (FailoverServerInfo)keyTrusteeClient.getServerInfo(targetMasterHostname);
    if (null == keyTrusteeServer) keyTrusteeServer = (FailoverServerInfo)keyTrusteeClient.getServerInfo();

    List<ConfigUpdate<?>> updates = new java.util.ArrayList<ConfigUpdate<?>>();
    updates.add(new ConfigUpdate<Boolean>(KEYTRUSTEE_SERVER_ROUND_ROBIN_KEY, keyTrusteeServer.isRoundRobinActive(),
            isRoundRobin(conf, env), "Client round robin request scheduling"));
    updates.add(new ConfigUpdate<Boolean>(KEYTRUSTEE_SERVER_INSECURE_KEY, keyTrusteeServer.isSslInsecure(),
            isSslInsecure(conf, env), "Client SSL insecurity"));
    updates.add(new ConfigUpdate<Integer>(KEYTRUSTEE_SERVER_HKP_PORT_KEY, keyTrusteeServer.getHkpPort(),
            getHkpPort(conf, env), "Client HKP port"));
    updates.add(new ConfigUpdate<Integer>(KEYTRUSTEE_SERVER_KTS_PORT_KEY, keyTrusteeServer.getKtsPort(),
            getKtsPort(conf, env), "Client KTS port"));
    updates.add(new ConfigUpdate<Boolean>(KEYTRUSTEE_SERVER_HKP_SSL_KEY, keyTrusteeServer.isSingleServerPort(),
            isHkpSsl(conf, env), "Client HKP SSL"));
    updates.add(new ConfigUpdate<String>(KEYTRUSTEE_SERVER_ACTIVE_HOST_KEY, keyTrusteeServer.getMasterHostname(),
            targetMasterHostname, "Active Hostname"));
    updates.add(new ConfigUpdate<List<String>>(KEYTRUSTEE_SERVER_PASSIVE_HOST_KEY, keyTrusteeServer.getHostnames(),
            getHostnames(conf, env), "Passive Hostname(s)", true));

    boolean processUpdates = forceUpdate;
    HashMap<String, ConfigUpdate<?>> updateMap = new HashMap<String, ConfigUpdate<?>>();

    // Don't overwrite config for no reason.
    // (This code could be removed along with the processUpdates flag above
    // However this will cause the config to be overwritten every time the
    // provider is instantiated).
    for (ConfigUpdate<?> update : updates) {
      if (update.updateRequired()) {
        if (LOG.isInfoEnabled()) LOG.info("Update detected for [" + update.getConfigKey() + "]");
        processUpdates = true;
      }
      updateMap.put(update.getConfigKey(), update);
    }

    if (processUpdates) {
      if (validUpdates(updateMap)) { // verifies active and passive KTS servers are both up and have matching identities
        FailoverServerInfo.Builder builder =
                (FailoverServerInfo.Builder) new FailoverServerInfo.Builder()
                        .hostnames((List<String>) updateMap.get(KEYTRUSTEE_SERVER_PASSIVE_HOST_KEY).getUpdate())
                        .setRoundRobin((Boolean) updateMap.get(KEYTRUSTEE_SERVER_ROUND_ROBIN_KEY).getUpdate())
                        .hostname((String) updateMap.get(KEYTRUSTEE_SERVER_ACTIVE_HOST_KEY).getUpdate())
                        .hkpPort((Integer) updateMap.get(KEYTRUSTEE_SERVER_HKP_PORT_KEY).getUpdate())
                        .ktsPort((Integer) updateMap.get(KEYTRUSTEE_SERVER_KTS_PORT_KEY).getUpdate())
                        .singleServerPort((Boolean) updateMap.get(KEYTRUSTEE_SERVER_HKP_SSL_KEY).getUpdate())
                        .certificate(keyTrusteeServer.getCertificatePath())
                        .sslInsecure((Boolean) updateMap.get(KEYTRUSTEE_SERVER_INSECURE_KEY).getUpdate())
                        .protocol(keyTrusteeServer.getProtocol()).tokenSync(keyTrusteeServer.isTokenSync())
                        .fingerprint(keyTrusteeServer.getFingerprint());

        FailoverServerInfo serverInfo = (FailoverServerInfo) builder.build();
        if (keyTrusteeServer.getHostname().equals(serverInfo.getHostname())) { // Minimal groundwork for supporting
                                                                               // server profiles. See OPSAPS-24133.
          keyTrusteeClient.removeServerInfo(keyTrusteeServer.getHostname());
        }
        keyTrusteeClient.putServerInfo(serverInfo);
        keyTrusteeClient.setDefaultServerInfo(serverInfo.getHostname());
        for (ConfigUpdate<?> update : updates) {
          if (LOG.isInfoEnabled()) LOG.info(update.getMessage());
        }
        return true;
      } else {
        if (LOG.isErrorEnabled()) LOG.error("Configuration update(s) failed due to previously logged warning(s) and/or error(s).");
        if (LOG.isWarnEnabled()) LOG.warn("Trustee Key Provider runtime configuration may not reflect Cloudera Manager settings.");
      }
    } else {
      if (LOG.isDebugEnabled()) LOG.debug("No configuration updates to process.");
    }
    return false;
  }
  
  /*
   * Verifies active and passive KTS servers are both up and have matching identities.
   */
  private static boolean validUpdates(Map<String, ConfigUpdate<?>> updateMap) throws IOException{
    boolean retVal = true;
    String[] hostUpdateCandidates = { KEYTRUSTEE_SERVER_ACTIVE_HOST_KEY, KEYTRUSTEE_SERVER_PASSIVE_HOST_KEY, 
            KEYTRUSTEE_SERVER_KTS_PORT_KEY, KEYTRUSTEE_SERVER_HKP_PORT_KEY, KEYTRUSTEE_SERVER_HKP_SSL_KEY, KEYTRUSTEE_SERVER_INSECURE_KEY };

    boolean hostUpdate = false;
    for (String hostUpdateCandidate : hostUpdateCandidates) {
      ConfigUpdate configUpdate = updateMap.get(hostUpdateCandidate);
      if (configUpdate.updateRequired()) hostUpdate = true;
    }
    if (hostUpdate) {
      String hkpScheme = "http";
      boolean hkpSecure = (Boolean) updateMap.get(KEYTRUSTEE_SERVER_HKP_SSL_KEY).getUpdate();
      if (hkpSecure) hkpScheme = "https";
      boolean insecureSsl = (Boolean) updateMap.get(KEYTRUSTEE_SERVER_INSECURE_KEY).getUpdate();
      String activeHost = (String) updateMap.get(KEYTRUSTEE_SERVER_ACTIVE_HOST_KEY).getUpdate();
      List<String> passiveHostList = (List<String>) updateMap.get(KEYTRUSTEE_SERVER_PASSIVE_HOST_KEY).getUpdate();
      int ktsPort = (Integer) updateMap.get(KEYTRUSTEE_SERVER_KTS_PORT_KEY).getUpdate();
      int hkpPort = (Integer) updateMap.get(KEYTRUSTEE_SERVER_HKP_PORT_KEY).getUpdate();

      java.util.Iterator<String> passiveHostIterator = passiveHostList.listIterator();

      URLConnector hkpConnector1 = new URLConnector(hkpScheme, activeHost, hkpPort, insecureSsl);
      URLConnector ktsConnector1 = new URLConnector("https", activeHost, ktsPort, insecureSsl);
      Connector.Response kts1Resp = null;
      try {
        kts1Resp = ktsConnector1.execQuery("/", "a=fingerprint", null);
      }catch (IOException e){
        retVal = false;
        if (LOG.isErrorEnabled()) {
          LOG.error("Fingerprint lookup failed on a KTS server: " + e.getMessage());
        }
        throw(e);
      }
      catch (Exception e) {
        retVal = false;
        if (LOG.isErrorEnabled()) {
          LOG.error("Fingerprint lookup failed on a KTS server: " + e.getMessage());
        }
      }

      while (passiveHostIterator.hasNext()) {
        String passiveHost = passiveHostIterator.next();
        if (!activeHost.equals(passiveHost)) {

          URLConnector hkpConnector2 = new URLConnector(hkpScheme, passiveHost, hkpPort, insecureSsl);
          URLConnector ktsConnector2 = new URLConnector("https", passiveHost, ktsPort, insecureSsl);

          Connector.Response kts2Resp = null;

          try {
            kts2Resp = ktsConnector2.execQuery("/", "a=fingerprint", null);
          } catch (IOException e) {
            retVal = false;
            if (LOG.isErrorEnabled()) {
              LOG.error("Fingerprint lookup failed on a KTS server: " + e.getMessage());
            }
            throw (e);
          } catch (Exception e) {
            retVal = false;
            if (LOG.isErrorEnabled()) {
              LOG.error("Fingerprint lookup failed on a KTS server: " + e.getMessage());
            }
          }

          String kts1Fingerprint = (kts1Resp != null) ? kts1Resp.getContent() : null;
          String kts2Fingerprint = (kts2Resp != null) ? kts2Resp.getContent() : null;

          if (null != kts1Fingerprint && null != kts2Fingerprint) {
            if (!kts1Fingerprint.equalsIgnoreCase(kts2Fingerprint)) {
              retVal = false;
              if (LOG.isWarnEnabled()) {
                LOG.warn("Fingerprint on active host [" + activeHost + " (" + kts1Fingerprint + ") ] does not equal " +
                        "fingerprint on passive host [" + passiveHost + " (" + kts2Fingerprint + ") ]");
              }
            }

            String hkpQueryFragment = "op=get&options=mr&search=0x";
            String rawFingerprint = kts1Fingerprint.substring(kts1Fingerprint.indexOf("/"));
            String hkpQuery = hkpQueryFragment + rawFingerprint;

            Connector.Response hkp1Resp = null, hkp2Resp = null;

            try {
              hkp1Resp = hkpConnector1.execQuery("/pks/lookup", hkpQuery, null);
              hkp2Resp = hkpConnector2.execQuery("/pks/lookup", hkpQuery, null);
            } catch (Exception e) {
              retVal = false;
              if (LOG.isErrorEnabled()) {
                LOG.error("Public key lookup failed on an HKP server: " + e.getMessage());
              }
            }

            String hkp1PublicKeyBlock = (hkp1Resp != null) ? hkp1Resp.getContent() : null;
            String hkp2PublicKeyBlock = (hkp2Resp != null) ? hkp2Resp.getContent() : null;

            if (null != hkp1PublicKeyBlock && null != hkp2PublicKeyBlock) {
              if (!hkp1PublicKeyBlock.equalsIgnoreCase(hkp2PublicKeyBlock)) {
                retVal = false;
                if (LOG.isWarnEnabled()) {
                  LOG.warn("Public key blocks on active and passive hosts do not match.");
                }
              }
            } else {
              LOG.warn("One or more KTS public key blocks [" + hkp1PublicKeyBlock + "] and [" + hkp2PublicKeyBlock + "] are null.");
            }
          } else {
            LOG.warn("One or more KTS fingerprints [" + kts1Fingerprint + "] and [" + kts2Fingerprint + "] are null.");
          }
        }
      }
    }
    return retVal;
  }

  private static URI[] splitUri(URI uri) throws URISyntaxException {
    String uriString = uri.toString();
    String[] uriStrings = uriString.split(",");
    URI[] uris = new URI[uriStrings.length];
    for (int i=0; i < uriStrings.length; i++) {
      uris[i] = new URI(uriStrings[i]);
    }
    return uris;
  }

  /*
   * Helper method for constructor only.
   */
  private void setKeyTrusteeConfDirAndURL(URI uri) throws MalformedURLException {
    if (this.env.containsKey(KEYTRUSTEE_CONF_DIR_ENV_VAR)) {
      this.keyTrusteeConfDir = new File(this.env.get(KEYTRUSTEE_CONF_DIR_ENV_VAR));
    } else {
      Preconditions.checkNotNull(uri,
              "Null uri passed to constructor and no" + KEYTRUSTEE_CONF_DIR_ENV_VAR + " environment " +
                      "variable set.");
      try {
        this.uris = splitUri(uri);
        Preconditions.checkState(uris.length > 0, "URI could not be processed: %s", uri);
        Path path = ProviderUtils.unnestUri(uris[0]);
        this.keyTrusteeConfDir = new File(path.toUri().toURL().getPath());
        this.keyTrusteeURL = path.toUri().toURL();
      } catch (URISyntaxException e) {
        throw new Error("URI could not be processed: " + uri);
      }
    }
  }

  /*
   * Helper method for constructor only.
   * Added in C5.7 (and may also be back-ported to C5.5.x releases after C5.7)
   * This method moves the default Key Trustee configuration directory from a symlinked location
   * to an absolute location.
   * The symlinked location is updated with each new parcel version, which has the effect of causing the
   * re-registration of the Key Trustee KMS, which can result in apparent or actual key loss.
   * See KT-2417 and TSB-121 in Cloudera Internal JIRA.
   */
  private void moveDefaultConfigurationDir() throws IOException {
    // If the C5.7 or C5.5.x KMS service is started with the  default KT config dir location (from the CSD),
    // the key provider code checks to see if a C5.5 KT config dir exists under the <PARCEL_ROOT>/KEYTRUSTEE-5.5...
    // If the parcel directory contains a non-empty /var/lib/kms-keytrustee/keytrustee/.keytrustee dir
    // AND the  default location (/var/lib/kms-keytrustee) is empty, the old config contents are copied to the
    // default location (/var/lib/kms-keytrustee) in order to retain the keys

    if (this.keyTrusteeConfDir.getAbsolutePath().equals(DEFAULT_KEYTRUSTEE_CONF_DIR_VALUE)) {
      if (emptyOrNonExistentDirectory(this.keyTrusteeConfDir)) {
        if (LOG.isInfoEnabled()) LOG.info("KT configuration directory is empty and default.");
        String parcelRoot = determineParcelPath();
        File oldParcel = new File(parcelRoot, DEFAULT_KEYTRUSTEE_55_PARCEL);
        if (oldParcel.exists() && oldParcel.isDirectory()) {
          if (LOG.isInfoEnabled()) LOG.info("Found existing parcel at " + oldParcel.getAbsolutePath() + ".");
          File oldConfigDir = new File(oldParcel, DEFAULT_KEYTRUSTEE_CONF_DIR_VALUE);
          if (oldConfigDir.exists() && oldConfigDir.isDirectory()) {
            if (LOG.isInfoEnabled()) LOG.info("Found existing KT configuration directory at " +
                    oldConfigDir.getAbsolutePath() + ".");
            if (FileUtils.sizeOfDirectory(oldConfigDir) > 0) {
              if (LOG.isInfoEnabled()) {
                LOG.info("Existing KT configuration directory at " +
                        oldConfigDir.getAbsolutePath() + " contains content.");
                LOG.info("Attempting copy of content from previous KT configuration directory (" +
                        oldConfigDir.getAbsolutePath() + ") to current KT configuration directory at (" +
                        keyTrusteeConfDir + ").");
                try {
                  // copy all readable files from old config dir to new config dir, preserving modification date
                  org.apache.commons.io.FileUtils.copyDirectory(oldConfigDir, this.keyTrusteeConfDir,
                          org.apache.commons.io.filefilter.CanReadFileFilter.CAN_READ, true);
                } catch (IOException e) {
                  String errorMessage = "Failure during copy of previous configuration directory contents " +
                          "to new configuration " +
                          "directory. Please contact Cloudera support for assistance before proceeding " +
                          "to avoid key loss.";
                  if (LOG.isErrorEnabled()) {
                    LOG.error(errorMessage, e);
                  }
                  throw new IOException(errorMessage, e);
                }
              }
            } else {
              if (LOG.isWarnEnabled()) LOG.warn("Existing KT configuration directory is empty.");
            }
          } else {
            if (LOG.isDebugEnabled()) LOG.debug("Found no existing KT configuration directory at " +
                    oldConfigDir.getAbsolutePath() + ".");
          }
        } else {
          if (LOG.isDebugEnabled()) LOG.debug("Found no parcel at " + oldParcel.getAbsolutePath() + ".");
        }
      } else {
        if (LOG.isInfoEnabled()) {
          LOG.info("Current configuration directory (" + this.keyTrusteeConfDir.getAbsolutePath() +
                  ") is not empty.");
          LOG.info("Content in current configuration directory will not be overwritten.");
        }
      }
    } else {
      if (LOG.isInfoEnabled()) {
        LOG.info("Skipping configuration move.");
        LOG.info("Current configuration directory: " + this.keyTrusteeConfDir.getAbsolutePath());
      }
    }
  }


  private String determineParcelPath() {
    String parcelPath = DEFAULT_PARCEL_LOCATION;
    String parcelRoot = System.getProperty("cdh.parcel.root"); // passed by CSD in control.sh
    if (LOG.isDebugEnabled()) LOG.debug("CDH PARCEL ROOT: " + parcelRoot);
    if (null != parcelRoot && (!parcelRoot.isEmpty())) {
      parcelPath = parcelRoot;
    }
    return new File(parcelPath).getAbsolutePath();
  }
  private boolean emptyOrNonExistentDirectory(File directory) {
    boolean returnValue = false;
    long sizeOfDir = -1;
    try {
      sizeOfDir =
              org.apache.commons.io.FileUtils.sizeOfDirectory(directory);
    } catch (IllegalArgumentException e) { // directory doesn't exist
      sizeOfDir = 0;
    } finally {
      if (sizeOfDir == 0) returnValue = true;
    }
    return returnValue;
  }

  public static int getKeyLength(Configuration conf, Environment env) {
    int keyLength = ClientFactory.DEFAULT_KEY_LENGTH.getLen();
    if (env.containsKey(KEYTRUSTEE_SERVER_KEY_LENGTH_ENV_VAR)) {
      keyLength = Integer.parseInt(env.get(KEYTRUSTEE_SERVER_KEY_LENGTH_ENV_VAR));
    } else {
      if (null != conf.get(KEYTRUSTEE_SERVER_KEY_LENGTH_KEY)) {
        keyLength = Integer.parseInt(conf.get(KEYTRUSTEE_SERVER_KEY_LENGTH_KEY));
      }
    }
    return keyLength;
  }

  public static boolean isUseKeyTrusteeConnectionPool(Configuration conf, Environment env) {
    boolean useConnectionPool = true;
    if (env.containsKey(KEYTRUSTEE_CLIENT_USE_POOL_ENV_VAR)) {
      useConnectionPool = Boolean.parseBoolean(env.get(KEYTRUSTEE_CLIENT_USE_POOL_ENV_VAR));
    } else {
      if (null != conf.get(KEYTRUSTEE_CLIENT_USE_POOL_KEY)) {
        useConnectionPool = Boolean.parseBoolean(conf.get(KEYTRUSTEE_CLIENT_USE_POOL_KEY));
      }
    }
    return useConnectionPool;
  }

  public static boolean isStrongLocking(Configuration conf, Environment env) {
    boolean strongLocking = false;
    if (env.containsKey(KEYTRUSTEE_CLIENT_STRONG_LOCKING_ENV_VAR)) {
      strongLocking = Boolean.parseBoolean(env.get(KEYTRUSTEE_CLIENT_STRONG_LOCKING_ENV_VAR));
    } else {
      if (null != conf.get(KEYTRUSTEE_CLIENT_STRONG_LOCKING_KEY)) {
        strongLocking = Boolean.parseBoolean(conf.get(KEYTRUSTEE_CLIENT_STRONG_LOCKING_KEY));
      }
    }
    return strongLocking;
  }

  public static boolean isSslInsecure(Configuration conf, Environment env) {
    boolean insecure = false;
    if (env.containsKey(KEYTRUSTEE_SERVER_INSECURE_ENV_VAR)) {
      insecure = Boolean.parseBoolean(env.get(KEYTRUSTEE_SERVER_INSECURE_ENV_VAR));
    } else {
      insecure = Boolean.parseBoolean(conf.get(KEYTRUSTEE_SERVER_INSECURE_KEY));
    }
    return insecure;
  }

  public static boolean isRoundRobin(Configuration conf, Environment env) {
    boolean roundRobin = false;
    if (env.containsKey(KEYTRUSTEE_SERVER_ROUND_ROBIN_ENV_VAR)) {
      roundRobin = Boolean.parseBoolean(env.get(KEYTRUSTEE_SERVER_ROUND_ROBIN_ENV_VAR));
    } else {
      roundRobin = Boolean.parseBoolean(conf.get(KEYTRUSTEE_SERVER_ROUND_ROBIN_KEY));
    }
    return roundRobin;
  }

  public static Integer getHkpPort(Configuration conf, Environment env)  {
    Integer hkpPort = null;
    try {
      if (env.containsKey(KEYTRUSTEE_SERVER_HKP_PORT_ENV_VAR)) {
        hkpPort = Integer.parseInt(env.get(KEYTRUSTEE_SERVER_HKP_PORT_ENV_VAR));
      } else {
        hkpPort = Integer.parseInt(conf.get(KEYTRUSTEE_SERVER_HKP_PORT_KEY));
      }
    } catch (NumberFormatException e) {
      hkpPort = null;
    }
    return hkpPort;
  }

  public static Integer getKtsPort(Configuration conf, Environment env) {
    Integer ktsPort = null;
    try {
      if (env.containsKey(KEYTRUSTEE_SERVER_KTS_PORT_ENV_VAR)) {
        ktsPort = Integer.parseInt(env.get(KEYTRUSTEE_SERVER_KTS_PORT_ENV_VAR));
      } else {
        ktsPort = Integer.parseInt(conf.get(KEYTRUSTEE_SERVER_KTS_PORT_KEY));
      }
    } catch (NumberFormatException e) {
      ktsPort = null;
    }
    return ktsPort;
  }

  public static Boolean isHkpSsl(Configuration conf, Environment env) {
    Boolean hkpSsl = null;
    if (env.containsKey(KEYTRUSTEE_SERVER_HKP_SSL_ENV_VAR)) {
      hkpSsl = Boolean.parseBoolean(env.get(KEYTRUSTEE_SERVER_HKP_SSL_ENV_VAR));
    } else {
      String hkpSslString = conf.get(KEYTRUSTEE_SERVER_HKP_SSL_KEY);
      if ((null == hkpSslString) || hkpSslString.equals("")) {
        hkpSsl = null;
      } else {
        hkpSsl = Boolean.parseBoolean(hkpSslString);
      }
    }
    return hkpSsl;
  }

  public static String getMasterHostname(Configuration conf, Environment env) {
    String hostname = null;
    if (env.containsKey(KEYTRUSTEE_SERVER_ACTIVE_HOST_ENV_VAR)) {
      hostname = env.get(KEYTRUSTEE_SERVER_ACTIVE_HOST_ENV_VAR);
    } else {
      hostname = conf.get(KEYTRUSTEE_SERVER_ACTIVE_HOST_KEY);
    }
    return hostname;
  }

  public static List<String> getHostnames(Configuration conf, Environment env) {
    List<String> hostnames = null;
    String masterHostname = null;
    if (env.containsKey(KEYTRUSTEE_SERVER_ACTIVE_HOST_ENV_VAR)) {
      masterHostname = env.get(KEYTRUSTEE_SERVER_ACTIVE_HOST_ENV_VAR);
    } else {
      masterHostname = conf.get(KEYTRUSTEE_SERVER_ACTIVE_HOST_KEY);
    }
    if (null == masterHostname || masterHostname.equals("")) {
      if (LOG.isWarnEnabled()) {
        LOG.warn("Master hostname is empty in configuration. Ignoring passive hostname(s) lacking validation criteria.");
      }
    } else {
      String passiveHostValue = null;
      if (env.containsKey(KEYTRUSTEE_SERVER_PASSIVE_HOST_ENV_VAR)) {
        passiveHostValue = env.get(KEYTRUSTEE_SERVER_PASSIVE_HOST_ENV_VAR);
      } else {
        passiveHostValue = conf.get(KEYTRUSTEE_SERVER_PASSIVE_HOST_KEY);
      }
      hostnames = Arrays.asList(rationalizeHostnames(passiveHostValue, masterHostname));
    }
    return hostnames;
  }

  static String[] rationalizeHostnames(String passiveHostValue, String hostname) {
    String[] hostnames = null;
    if (null != passiveHostValue && (!passiveHostValue.equals(""))) {
      hostnames = passiveHostValue.split(",");
      boolean contained = false;
      int containedIndex = 0;
      for (int i=0; i < hostnames.length; i++) {
        if (hostnames[i].equalsIgnoreCase(hostname)) {
          contained = true;
          containedIndex = i;
        }
      }
      if (!contained) {
        List<String> hostList = new java.util.ArrayList(Arrays.asList(hostnames));
        hostList.add(0, hostname);
        hostnames = hostList.toArray(new String[hostList.size()]);
      } else {
        if (containedIndex != 0) {
          List<String> hostList = new java.util.ArrayList(Arrays.asList(hostnames));
          hostList.remove(hostname);
          hostList.add(0, hostname);
          hostnames = hostList.toArray(new String[hostList.size()]);
        }
      }
    } else {
      hostnames = new String[1];
      hostnames[0] = hostname;
    }
    return hostnames;
  }

  public static int getMaxIdle(Configuration conf, Environment env) {
    return getIntValueFromConfigOrEnv(conf, env,
            KEYTRUSTEE_CLIENT_POOL_MAX_IDLE_KEY, KEYTRUSTEE_CLIENT_POOL_MAX_IDLE_ENV_VAR,
            PooledClientConnectionFactory.DEFAULT_MAX_IDLE);
  }

  public static int getMinIdle(Configuration conf, Environment env) {
    return getIntValueFromConfigOrEnv(conf, env,
            KEYTRUSTEE_CLIENT_POOL_MIN_IDLE_KEY, KEYTRUSTEE_CLIENT_POOL_MIN_IDLE_ENV_VAR,
            PooledClientConnectionFactory.DEFAULT_MIN_IDLE);
  }

  public static int getMaxTotal(Configuration conf, Environment env) {
    return getIntValueFromConfigOrEnv(conf, env,
            KEYTRUSTEE_CLIENT_POOL_MAX_SIZE_KEY, KEYTRUSTEE_CLIENT_POOL_MAX_SIZE_ENV_VAR,
            PooledClientConnectionFactory.DEFAULT_MAX_SIZE);
  }

  public static int getAbandonedTimeout(Configuration conf, Environment env) {
    return getIntValueFromConfigOrEnv(conf, env,
            KEYTRUSTEE_CLIENT_POOL_ABANDONED_TIMEOUT_KEY, KEYTRUSTEE_CLIENT_POOL_ABANDONED_TIMEOUT_ENV_VAR,
            PooledClientConnectionFactory.DEFAULT_ABANDONED_TIMEOUT);
  }

  public static int getHttpReadConnectionTimeout(Configuration conf, Environment env) {
    return getIntValueFromConfigOrEnv(conf, env,
            KEYTRUSTEE_CLIENT_READ_CONNECTION_TIMEOUT_KEY, KEYTRUSTEE_CLIENT_READ_CONNECTION_TIMEOUT_ENV_VAR,
            com.cloudera.keytrustee.URLConnector.DEFAULT_TIMEOUT);
  }

  private static int getIntValueFromConfigOrEnv(Configuration conf, Environment env, String confKeyName, String envKeyName, int defaultValue) {
    int retVal = defaultValue;
    if (env.containsKey(envKeyName)) {
      retVal = Integer.parseInt(env.get(envKeyName));
    } else {
      if (null != conf.get(confKeyName)) {
        retVal = Integer.parseInt(conf.get(confKeyName));
      }
    }
    return retVal;
  }

  public final class Register {
    private Configuration conf;
    private Environment env;
    private String hostname;
    private String[] hostnames;
    private String certPath;
    private String protocol;
    private String org;
    private String auth;
    private boolean insecure;
    private boolean tokenSync;
    private boolean roundRobin;
    private Integer hkpPort;
    private Integer ktsPort;
    private Boolean hkpSsl;
    private boolean failed = false;

    public Register(Configuration conf, Environment env) {
      this.conf = conf;
      this.env = env;
    }

    public String getHostname() {
      return hostname;
    }

    public String[] getHostnames() { return hostnames; }

    public String getCertPath() {
      return certPath;
    }

    public String getProtocol() {
      return protocol;
    }

    public String getOrg() {
      return org;
    }

    public String getAuth() {
      return auth;
    }

    public boolean isInsecure() {
      return insecure;
    }

    public boolean isTokenSync() {
      return tokenSync;
    }

    public boolean isRoundRobin() { return roundRobin; }

    public Integer getHkpPort() { return hkpPort; }

    public Integer getKtsPort() { return ktsPort; }

    public Boolean getHkpSsl() { return hkpSsl; }

    public Register invoke() {
      if (env.containsKey(KEYTRUSTEE_SERVER_ACTIVE_HOST_ENV_VAR)) {
        hostname = env.get(KEYTRUSTEE_SERVER_ACTIVE_HOST_ENV_VAR);
      } else {
        hostname = conf.get(KEYTRUSTEE_SERVER_ACTIVE_HOST_KEY);
      }

      String passiveHostValue = null;
      if (env.containsKey(KEYTRUSTEE_SERVER_PASSIVE_HOST_ENV_VAR)) {
        passiveHostValue = env.get(KEYTRUSTEE_SERVER_PASSIVE_HOST_ENV_VAR);
      } else {
        passiveHostValue = conf.get(KEYTRUSTEE_SERVER_PASSIVE_HOST_KEY);
      }
      rationalizeHostnames(passiveHostValue);

      if (env.containsKey(KEYTRUSTEE_SERVER_CERT_FILE_ENV_VAR)) {
        certPath = env.get(KEYTRUSTEE_SERVER_CERT_FILE_ENV_VAR);
      } else {
        certPath = conf.get(KEYTRUSTEE_SERVER_CERT_FILE_KEY);
        if (null !=certPath && certPath.equals("")) certPath = null;
      }

      if (env.containsKey(KEYTRUSTEE_SERVER_PROTOCOL_ENV_VAR)) {
        protocol = env.get(KEYTRUSTEE_SERVER_PROTOCOL_ENV_VAR);
      } else {
        protocol = conf.get(KEYTRUSTEE_SERVER_PROTOCOL_KEY);
        if (null == protocol || protocol.equals("")) {
          protocol = ServerInfo.Builder.DEFAULT_PROTOCOL;
        }
      }

      if (env.containsKey(KEYTRUSTEE_SERVER_ORG_ENV_VAR)) {
        org = env.get(KEYTRUSTEE_SERVER_ORG_ENV_VAR);
      } else {
        org = conf.get(KEYTRUSTEE_SERVER_ORG_KEY);
      }

      if (env.containsKey(KEYTRUSTEE_SERVER_AUTH_ENV_VAR)) {
        auth = env.get(KEYTRUSTEE_SERVER_AUTH_ENV_VAR);
      } else {
        auth = conf.get(KEYTRUSTEE_SERVER_AUTH_KEY);
      }

      if (env.containsKey(KEYTRUSTEE_SERVER_INSECURE_ENV_VAR)) {
        insecure = Boolean.parseBoolean(env.get(KEYTRUSTEE_SERVER_INSECURE_ENV_VAR));
      } else {
        insecure = Boolean.parseBoolean(conf.get(KEYTRUSTEE_SERVER_INSECURE_KEY));
      }

      if (env.containsKey(KEYTRUSTEE_SERVER_TOKEN_SYNC_ENV_VAR)) {
        tokenSync = Boolean.parseBoolean(env.get(KEYTRUSTEE_SERVER_TOKEN_SYNC_ENV_VAR));
      } else {
        tokenSync = Boolean.parseBoolean(conf.get(KEYTRUSTEE_SERVER_TOKEN_SYNC_KEY));
      }

      if(env.containsKey(KEYTRUSTEE_SERVER_ROUND_ROBIN_ENV_VAR)) {
        roundRobin = Boolean.parseBoolean(env.get(KEYTRUSTEE_SERVER_ROUND_ROBIN_ENV_VAR));
      } else {
        roundRobin = Boolean.parseBoolean(conf.get(KEYTRUSTEE_SERVER_ROUND_ROBIN_KEY));
      }

      try {
        if (env.containsKey(KEYTRUSTEE_SERVER_HKP_PORT_ENV_VAR)) {
          hkpPort = Integer.parseInt(env.get(KEYTRUSTEE_SERVER_HKP_PORT_ENV_VAR));
        } else {
          hkpPort = Integer.parseInt(conf.get(KEYTRUSTEE_SERVER_HKP_PORT_KEY));
        }
      } catch (NumberFormatException e) {
        hkpPort = null;
        if (LOG.isDebugEnabled()) {
          LOG.debug(e);
        }
        if (LOG.isWarnEnabled()) {
          LOG.warn("Could not find and/or parse HKP port in configuration. " +
                  "Setting hkpPort to null. Expect auto-negotiation with server.");
        }
      }
      try {
        if (env.containsKey(KEYTRUSTEE_SERVER_KTS_PORT_ENV_VAR)) {
          ktsPort = Integer.parseInt(env.get(KEYTRUSTEE_SERVER_KTS_PORT_ENV_VAR));
        } else {
          ktsPort = Integer.parseInt(conf.get(KEYTRUSTEE_SERVER_KTS_PORT_KEY));
        }
      } catch (NumberFormatException e) {
        ktsPort = null;
        if (LOG.isDebugEnabled()) {
          LOG.debug(e);
        }
        if (LOG.isWarnEnabled()) {
          LOG.warn("Could not find and/or parse KTS port in configuration. Setting ktsPort to null. " +
                  "Expect auto-negotiation with server.");
        }
      }

      if(env.containsKey(KEYTRUSTEE_SERVER_HKP_SSL_ENV_VAR)) {
        hkpSsl = Boolean.parseBoolean(env.get(KEYTRUSTEE_SERVER_HKP_SSL_ENV_VAR));
      } else {
        hkpSsl = Boolean.parseBoolean(conf.get(KEYTRUSTEE_SERVER_HKP_SSL_KEY));
      }
      if (null == hkpSsl) {
        // hkpSsl = null;
        if (LOG.isWarnEnabled()) {
          LOG.warn("Could not find HKP SSL setting in configuration. Setting hkpSsl to null. " +
                  "Expect auto-negotiation with server.");
        }
      }

      return this;
    }

    private String getFromEnvironmentOrConfiguration(String environmentVariable, String configKey) {
      if (env.containsKey(environmentVariable)) {
        return env.get(environmentVariable);
      } else {
        return conf.get(configKey);
      }
    }

    private void rationalizeHostnames(String passiveHostValue) {
      hostnames = TrusteeKeyProviderConfiguration.rationalizeHostnames(passiveHostValue, hostname);
    }

    public void invariant() throws IOException {
      this.failed = false;
      String message = "TrusteeKeyProvider registration aborted: \n";
      if (empty(hostname)) { message = fail(message,
              KEYTRUSTEE_SERVER_ACTIVE_HOST_ENV_VAR, KEYTRUSTEE_SERVER_ACTIVE_HOST_KEY); }
      if (empty(certPath)) {
        if (LOG.isWarnEnabled()) {
          LOG.warn("No KeyTrustee key provider certPath specified. Using JVM default trust store.");
        }
        //message = fail(message, KEYTRUSTEE_SERVER_CERT_FILE_ENV_VAR, KEYTRUSTEE_SERVER_CERT_FILE_KEY);
      }
      //if (empty(protocol)) { message = fail(message,
      //        KEYTRUSTEE_SERVER_PROTOCOL_ENV_VAR, KEYTRUSTEE_SERVER_PROTOCOL_KEY); }
      if (empty(org)) { message = fail(message,
              KEYTRUSTEE_SERVER_ORG_ENV_VAR, KEYTRUSTEE_SERVER_ORG_KEY); }
      if (empty(auth)) { message = fail(message,
              KEYTRUSTEE_SERVER_AUTH_ENV_VAR, KEYTRUSTEE_SERVER_AUTH_KEY); }
      if (failed) {
        throw new IOException(message);
      }
    }

    private boolean empty(String candidate) {
      if (null == candidate || candidate.equals("")) return true;
      return false;
    }

    private String fail(String currentMessage, String envVar, String confKey) {
      this.failed = true;
      String message = currentMessage + "Set " + envVar +
              " environment variable OR " + confKey + " key in Hadoop KMS configuration file. \n";
      return message;
    }

  }

}
